import json
import pathlib
from typing import Dict, List, Union

import requests
from loguru import logger as LOGGER


class ApiTokenHelper():
    """
    Manage dt_tools* 3rd Party API interface tokens.

    """
    _DT_TOOLS_TOKENS_LOCATION=pathlib.Path('~').expanduser().absolute() / ".dt_tools" / "api_tokens.json"
    
    API_IP_INFO = 'ipinfo.io'
    API_WEATHER_INFO = 'weatherapi.com'
    API_GEOLOCATION_INFO = 'geocode.maps.co'
    
    _API_DICT = {
        "ipinfo.io": {
            "desc": "IP Address information API",
            "package": "dt-net",
            "module": "dt_tools.net.ip_helper",
            "token_url": "https://ipinfo.io/missingauth",
            "validate_url": "https://ipinfo.io/8.8.8.8?token={token}",
            "limits": "50,000/month, ~1,600/day"
        },
        "weatherapi.com": {
            "desc": "Weather API (current, forecasts, alerts)",
            "package": "dt-misc",
            "module": "dt_tools.misc.weather",
            "token_url": "https://www.weatherapi.com/signup.aspx",
            "validate_url": "http://api.weatherapi.com/v1/ip.json?key={token}&q=auto:ip",
            "limits": "1,000,000/month, ~32,000/day"
        },
        "geocode.maps.co": {
            "desc": "GeoLocation API (Lat, Lon, Address, ...)",
            "package": "dt-misc",
            "module": "dt_tools.misc.geoloc",
            "token_url": "https://geocode.maps.co/join/",
            "validate_url": "https://geocode.maps.co/reverse?lat=0&lon=0&api_key={token}",
            "limits": "5,000/day, throttle 1 per sec"
        }
    }
    
    @classmethod
    def _get_tokens_dictionary(cls) -> Dict[str, dict]:
        cls._DT_TOOLS_TOKENS_LOCATION.parent.mkdir(parents=True, exist_ok=True)
        token_dict = {}
        if cls._DT_TOOLS_TOKENS_LOCATION.exists():
            token_dict = json.loads(cls._DT_TOOLS_TOKENS_LOCATION.read_text())
        return token_dict
    
    @classmethod
    def get_api_token(cls, service_id: str) -> str:
        """
        Get token for API service_id

        Args:
            service_id (str): Service identifier (see get_api_services())

        Raises:
            NameError: If the service name is not valid.

        Returns:
            str: API token for target service or None if not in token cache.
        """
        if cls._API_DICT.get(service_id, None) is None:
            raise NameError(f'Not a valid service: {service_id}')
        
        t_dict = cls._get_tokens_dictionary()
        token = t_dict.get(service_id, None)
        return token

    @classmethod
    def save_api_token(cls, service_id: str, token: str) -> bool:
        """
        Save the API Token for the service.

        Args:
            service_id (str): Target service id.
            token (str): Token string.

        Returns:
            bool: True if saved, False if there was an error.
        """
        saved = True
        t_dict = cls._get_tokens_dictionary()
        t_dict[service_id] = token
        token_str = json.dumps(t_dict)
        try:
            cls._DT_TOOLS_TOKENS_LOCATION.write_text(token_str)
        except Exception as ex:
            LOGGER.error(f'Unable to save token for {service_id} - {repr(ex)}')
            saved = False

        return saved
    
    @classmethod
    def get_api_service_ids(cls) -> List[str]:
        """
        Return a list of the API service ids.

        Returns:
            List[str]: List of API service ids.
        """
        t_dict = cls._get_tokens_dictionary()
        return list(t_dict.keys())

    @classmethod
    def get_api_service_definition(cls, service_id: str) -> Union[Dict, None]:
        """
        Return a dictionary of the API Service.

        Args:
            api_key (str): Service ID of requested service.

        Returns:

            Union[Dict, None]: Service definition as a dict if found, else None.

            Format:: 
            
                <service_id1>: {
                    "desc": "Service description",
                    "package": "dt-xxxxx",
                    "module": "dt_tools.xxx.xxxx",
                    "token_url": "https://xxxxxx",
                    "validate_url": "https://xxxxxx/yyyyy",
                    "limits": "Limit description",
                    "enabled": True
                }

        """
        return cls._API_DICT.get(service_id, None)
    
    @classmethod
    def can_validate(cls, service: str) -> bool:
        """
        Ensure service setup is valid.

        Args:
            service_id (str): Target service_id.

        Returns:
            bool: True if service is setup False if invalid name or missing token.
        """
        t_service:dict = cls._API_DICT.get(service, {})
        valid_service = t_service.get('validate_url', None) is not None
        if valid_service:
            LOGGER.debug(f'{service} is a valid service name')
            valid_service = cls.get_api_token(service) is not None
            if valid_service:
                LOGGER.debug(f'{service} has a token.')
            else:
                LOGGER.debug(f'{service} does NOT have a token.')
        else:
            LOGGER.debug(f'{service} is NOT a valid service name')

        return valid_service

    @classmethod
    def validate_token(cls, service_id: str) -> bool:
        """
        Validate service token.

        Args:
            service_id (str): Target service_id.

        Returns:
            bool: True if token validates successfully, False if invalid token.
        """
        valid_token = False
        if not cls.can_validate(service_id):
            LOGGER.debug(f'{service_id} is not valid.')
        else:
            entry = cls.get_api_service_definition(service_id)
            try:
                token = cls.get_api_token(service_id)
                url = entry['validate_url'].replace('{token}', token)
                resp = requests.get(url)
                LOGGER.debug(f'Validate: {url}  returns: {resp.status_code}')
                if resp.status_code == 200:
                    valid_token = True
            except Exception as ex:
                LOGGER.debug(f'Validate: {url}  Exception: {repr(ex)}')

        return valid_token
    


"""
This module creates the token file and stores token(s) used for interface with dt_tools 3rd party entities.

Specifically, tokens required for the follwing services:

    - ipinfo.io
    - weatherapi.com
    - geocode.maps.co

```
poetry run python -m dt_tools.cli.set_api_tokens_cli.py
```

"""
import argparse
import sys
from time import sleep
from typing import List

from loguru import logger as LOGGER

import dt_tools.logger.logging_helper as lh
from dt_tools.misc.api_helper import ApiTokenHelper as api_helper
from dt_tools.misc.helpers import ObjectHelper as oh
from dt_tools.os.project_helper import ProjectHelper


def get_input(text: str, valid_responses: List=None) -> str:
    resp = input(text)
    if valid_responses is not None:
        while resp not in valid_responses:
            resp = input(text)
    return resp

def list_apis():
    LOGGER.info('API Services available')
    LOGGER.info('-----------------------------------------')
    log_level = ''
    idx = 0
    for key, val in api_helper._API_DICT.items():
        entry = oh.dict_to_obj(val)
        has_token = api_helper.get_api_token(key) is not None
        valid = ""
        if has_token:
            valid_token = api_helper.validate_token(key)
            log_level = 'SUCCESS' if valid_token else 'ERROR'
            valid = "(Valid token)" if valid_token else "(Invalid token)"
        else:
            log_level = 'WARNING'
        LOGGER.log(log_level, f'{idx:1} {key} - {entry.desc}')
        LOGGER.log(log_level, f'  Has Token    : {has_token} {valid}')
        LOGGER.info(f'  Sign-up  URL : {entry.token_url}')
        LOGGER.info(f'  Validate URL : {entry.validate_url}')
        LOGGER.info(f'  Limits       : {entry.limits}')
        LOGGER.info(f'  dt_module    : {entry.module}')
        LOGGER.info('')
        idx += 1

def select_api() -> str:
    num_services = len(api_helper._API_DICT)
    choices = []
    for i in range(num_services):
        choices.append(str(i))
    choices.append('99')

    resp = int(get_input(f'What service to set API for? (0-{num_services-1} or 99 to exit) > ', valid_responses=choices))
    key = None
    if resp != 99:
        key = list(api_helper._API_DICT.keys())[resp]

    return key

def manage_token(api_key: str) -> str:
    rc = 0
    entry = api_helper.get_api_service_definition(api_key)
    if entry is None:
        raise ValueError(f'Unknown API service: {api_key}')
    LOGGER.info('')
    LOGGER.info('-'*90)
    LOGGER.info('')
    LOGGER.info(f'Service  : {api_key} - {entry["desc"]}')
    LOGGER.info(f'Token URL: {entry["token_url"]}')
    LOGGER.info('')
    LOGGER.warning('NOTE:')
    LOGGER.info(f'  The token is stored locally in {api_helper._DT_TOOLS_TOKENS_LOCATION}.')
    LOGGER.info(f'         format: {{"{api_key}": "xxxxxxxxxxxxxx"}}')
    LOGGER.info('')

    old_token = api_helper.get_api_token(api_key) 
    if old_token is None:
        prompt = 'Continue (y/n) > '
    else:
        prompt = 'Token exists, overwrite? > '

    if get_input(prompt, ['y', 'n']) == 'y':
        token = get_input('Token > ')
        if len(token.strip()) == 0:
            LOGGER.warning('  Empty token, did not save.')
            rc = 2
        else:
            if api_helper.save_api_token(api_key, token):
                if api_helper.can_validate(api_key):
                    if api_helper.validate_token(api_key):
                        LOGGER.success('Token saved.')
                    else:
                        if old_token is None:
                            api_helper.save_api_token(api_key, None)
                            rc = 3
                        else:
                            api_helper.save_api_token(api_key, old_token)
                            rc = 4
                        LOGGER.warning(f'Token not valid, not saved. ({rc})')
        LOGGER.info('')
    
    return rc

def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('-l', '--list', action='store_true', default=False, help='List API information.')
    parser.add_argument('-v', '--verbose', action='store_true', default=False, help='Verbose (debug) logging.')
    args = parser.parse_args()
    log_level = 'DEBUG' if args.verbose else 'INFO'
    lh.configure_logger(log_level=log_level, brightness=False)

    rc = 0
    version = ProjectHelper.determine_version('dt-foundation')
    LOGGER.info('')
    LOGGER.info('-'*95)
    LOGGER.info(f' dt_tools Token Manager (v{version})')
    LOGGER.info('-'*95)
    LOGGER.info('')
    LOGGER.info('To enable the dt_tools and packages for a specific service, a one-time process is necessary')
    LOGGER.info('to aquire a FREE API token.  Once aquired, this process will save it locally for future use.')
    LOGGER.info('')
    LOGGER.info('To get a token for a specific service, go the the sign-up URL (below) and follow the process')
    LOGGER.info('for creating an API token.  Provide the token here and it will be cached.')
    LOGGER.info('')
    LOGGER.info('If you already have a token, but forget what it is, you may log back into the service provider')
    LOGGER.info('and retrieve your token.')
    LOGGER.info('')
    LOGGER.info('-'*95)
    sleep(3)

    if args.list:
        list_apis()
    else:
        list_apis()
        api_key = select_api()
        if api_key is None:
            LOGGER.warning('  No API selected.')
            rc = 1
        else:
            rc = manage_token(api_key)

    return rc


if __name__ == "__main__":
    sys.exit(main())
    # manage_token()



"""
This module demonstrates the OSHelper features.

- OS Detection
- Elevated privilege detection and escalation

"""
from loguru import logger as LOGGER

import dt_tools.logger.logging_helper as lh
from dt_tools.os.os_helper import OSHelper
from dt_tools.misc.helpers import ObjectHelper as oh
from types import SimpleNamespace
import json

def demo_simple_report(sysinfo: dict):
    for section, section_value in sysinfo.items():
        LOGGER.success(section)
        for k,v in section_value.items():
            if isinstance(v, list):
                instance_no = 0
                for l_entry in v:
                    LOGGER.success(f'  == {instance_no} ===')
                    instance_no += 1
                    for k2,v2 in l_entry.items():
                        LOGGER.info(f'    {k2:10} {v2}')
            else:
                LOGGER.info(f'  {k:18} : {v}')
        LOGGER.info('')

def demo_report(sysinfo: dict):
    obj = oh.dict_to_obj(sysinfo)
    LOGGER.debug(obj.system)
    LOGGER.info('')
    LOGGER.success('System Info:')
    LOGGER.info(f'  Host      : {obj.system.hostname} - {obj.system.host_fqdn}')
    LOGGER.info(f'              {obj.system.ip}')
    LOGGER.info(f'  Platform  : {obj.system.platform} - {obj.system.machine_type}')
    LOGGER.info(f'  OS        : {obj.system.os}')
    LOGGER.info(f'              {obj.system.os_ver}')
    LOGGER.info(f'  Kernel    : {obj.system.os_kernel}')
    LOGGER.info(f'  Boot Info : Last boot - {obj.system.last_boot_time}')
    LOGGER.info(f'              Up time   - {obj.system.uptime}')
    
    LOGGER.info('')
    LOGGER.debug(obj.cpu)
    LOGGER.success('CPU Info:')    
    LOGGER.info(f'  Processor : {obj.cpu.processor}')
    LOGGER.info(f'  Cores     : {obj.cpu.cores_physical:2} physical')
    LOGGER.info(f'              {obj.cpu.cores_logical:2} logical')
    LOGGER.info(f'  Frequency : {obj.cpu.freq_min} - {obj.cpu.freq_max} MHz')

    LOGGER.info('')
    LOGGER.debug(obj.memory)
    LOGGER.success('Memory Info:')
    LOGGER.info( '               Total       Used       Free  % Used')
    LOGGER.info( '           ---------  ---------  ---------  ------')
    total = OSHelper.bytes_to_printformat(obj.memory.virtual_total)
    used  = OSHelper.bytes_to_printformat(obj.memory.virtual_used)
    free  = OSHelper.bytes_to_printformat(obj.memory.virtual_free)
    pct_used = f'{obj.memory.virtual_pct_used}%'
    LOGGER.info(f'  Virtual  {total:>9}  {used:>9}  {free:>9}  {pct_used:>6}')
    total = OSHelper.bytes_to_printformat(obj.memory.swap_total)
    used  = OSHelper.bytes_to_printformat(obj.memory.swap_used)
    free  = OSHelper.bytes_to_printformat(obj.memory.swap_free)
    pct_used = f'{obj.memory.swap_pct_used}%'
    LOGGER.info(f'  Swap     {total:>9}  {used:>9}  {free:>9}  {pct_used:>6}')
    
    LOGGER.info('')
    LOGGER.success('Disk Info:')
    print('  Device       Type            Total      Used       Free    % Used')
    print('  ------------ ---------- ---------- ---------- ---------- --------')
    for de in obj.disk.partitions:
        # print(de)
        total = '-'
        used = '-'
        free = '-'
        used_pct = '-'
        type = de.fstype if len(de.fstype) > 0 else de.mount_opts
        if hasattr(de, 'total'):
            total = OSHelper.bytes_to_printformat(de.total)
            used = OSHelper.bytes_to_printformat(de.used)
            free = OSHelper.bytes_to_printformat(de.free)
            used_pct = f'{de.used_pct}%'
        print(f'  {de.device:12} {type:10} {total:>10} {used:>10} {free:>10} {used_pct:>8}')
    

def demo():
    LOGGER.info('')
    LOGGER.info('-'*40)
    LOGGER.info('dt_misc_sysinfo_demo')
    LOGGER.info('-'*40)
    sysinfo = OSHelper.sysinfo(include_all=True)

    LOGGER.info('')
    LOGGER.info('system_information (as dictionary):')
    print(json.dumps(sysinfo, indent=1))
    input('\nPress Enter to continue... ')

    LOGGER.info('')
    LOGGER.info('system_information (simple report):')
    demo_simple_report(sysinfo)
    LOGGER.info('')
    input('\nPress Enter to continue... ')

    LOGGER.info('')
    LOGGER.info('system_information (as object):')
    sysinfo_obj:SimpleNamespace  = oh.dict_to_obj(sysinfo)
    LOGGER.success('sysinfo_obj.system:')
    LOGGER.info(sysinfo_obj.system)
    LOGGER.success('sysinfo_obj.cpu:')
    LOGGER.info(sysinfo_obj.cpu)
    LOGGER.success('sysinfo_obj.memory:')
    LOGGER.info(sysinfo_obj.memory)
    LOGGER.success('sysinfo_obj.disk:')
    LOGGER.info(sysinfo_obj.disk)

    LOGGER.info('')
    LOGGER.info('system_information (as report):')
    demo_report(sysinfo)

    LOGGER.info('')
    LOGGER.info('Demo commplete.')            
    input('\nPress Enter to continue... ')

if __name__ == "__main__":
    lh.configure_logger(log_format=lh.DEFAULT_CONSOLE_LOGFMT, log_level="INFO", brightness=False)
    demo()



"""
Logging helper methods for loguru. (https://github.com/Delgan/loguru)

Example::

    import dt_tools.logger.logging_helper as lh
    from loguru import logger as LOGGER

    log_file = './mylog.log'

    # File logger
    f_handle = lh.configure_logger(log_target=log_file, log_level="DEBUG")
    # Console logger
    c_handle - lh.configure_logger()

    LOGGER.debug('this should only show up in file logger')
    LOGGER.info('this should show up in file logger and console')

"""
import functools
import inspect
import logging
import sys
import time
from typing import Dict, List

from loguru import logger as LOGGER

# Format variables:
# Variable    Description
# ----------- ------------------------------------------------------------
# elapsed     The time elapsed since the start of the program
# exception   The formatted exception if any, None otherwise
# extra       The dict of attributes bound by the user (see bind())
# file        The file where the logging call was made
# function    The function from which the logging call was made
# level       The severity used to log the message
# line        The line number in the source code
# message     The logged message (not yet formatted)
# module      The module where the logging call was made
# name        The __name__ where the logging call was made
# process     The process in which the logging call was made
# thread      The thread in which the logging call was made
# time        The aware local time when the logging call was made

DEFAULT_FILE_LOGFMT = "<green>{time:MM/DD/YY HH:mm:ss}</green> |<level>{level: <8}</level>|<cyan>{file:18}</cyan>|<cyan>{line:4}</cyan>| <level>{message}</level>"
"""For file logging, format- timestamp \|level\|method name\|lineno\|message"""

DEFAULT_CONSOLE_LOGFMT = "<level>{message}</level>"
"""For console logging, format- message"""

DEFAULT_DEBUG_LOGFMT =  "<green>{time:HH:mm:ss}</green> |<level>{level: <8}</level>|<cyan>{module:20}</cyan>|<cyan>{line:4}</cyan>| <level>{message}</level>"
"""For console/file logging, timestamp \|level\|method name\|lineno\|message"""

DEFAULT_DEBUG_LOGFMT2 =  "<green>{time:HH:mm:ss}</green> |<level>{level: <8}</level>|<yellow>{name:15}|{module:20}|{line:4}</yellow>| <level>{message}</level>"
"""For console/file logging, timestamp \|level\|method name\|lineno\|message"""

def configure_logger(log_target = sys.stderr, 
                     log_level: str = "INFO", 
                     log_format: str = None, 
                     brightness: bool = False, 
                     log_handle: int = 0, 
                     enqueue: bool = False,
                     propogate_loggers: List[str] = None,
                     enable_loggers: List[str] = None,
                     disable_loggers: List[str] = None,
                     **kwargs) -> int:
    """
    Configure logger via loguru.

     - should be done once for each logger (console, file,..)
     - if reconfiguring a logger, pass the log_handle
    
    Parameters:
        log_target: defaults to stderr, but can supply filename as well (default console/stderr)
        log_level : TRACE|DEBUG|INFO(dflt)|ERROR|CRITICAL (default INFO)
        log_format: format for output log line (loguru default)
        brightness: console messages bright or dim (default False)
        log_handle: handle of log being re-initialized. (default 0)
        enqueue   : to avoid collisions on the log in muti-processing applications. 
            See is_complete() and wait_for_complete().
        propogate_loggers: list of (legacy) loggers to propogate this configuration to.
        enable_loggers: list of loggers to enable messages from
        disable_loggers: list of loggers to disable messages from
        other     : keyword args related to loguru logger.add() function
            see: https://loguru.readthedocs.io/en/stable/api/logger.html#loguru._logger.Logger.add            

    Example::

        import dt_tools.logger.logging_helper as lh
        from loguru import logger as LOGGER

        log_file = './mylog.log'

        f_handle = lh.configure_logger(log_target=log_file, log_level="DEBUG")
        c_handle - lh.configure_logger()

        LOGGER.debug('this should only show up in file logger')
        LOGGER.info('this should show up in file logger and console')

    Returns:
        logger_handle_id: integer representing logger handle
    """
    # Ideas from: https://medium.com/@muh.bazm/how-i-unified-logging-in-fastapi-with-uvicorn-and-loguru-6813058c48fc
    try:
        if log_handle >= 0:
            # Remove specific handler
            LOGGER.remove(log_handle)
            LOGGER.trace(f'removed handler: {log_handle}')            
    except Exception as ex:
        if 'There is no existing' not in str(ex):
            LOGGER.error(f'configure_logger(): {ex}')

    # Intercept standard logging
    logging.basicConfig(handlers=[_InterceptHandler()], level=0, force=True)    

    if brightness is not None:
        set_log_levels_brightness(brightness)
        
    if log_format is None:
        # Set format based on type of logger (file vs console)        
        log_format = DEFAULT_FILE_LOGFMT if isinstance(log_target, str) else DEFAULT_CONSOLE_LOGFMT

    hndl = LOGGER.add(sink=log_target, 
                      level=log_level, 
                      format=log_format, 
                      enqueue=enqueue,
                      diagnose=True, 
                      **kwargs)
    # NOTE: This adds a new variable to LOGGER so current log level can be determined.
    LOGGER.log_level = log_level

    if propogate_loggers is not None:
        _propogate_loggers(propogate_loggers)
    if enable_loggers is not None:
        _enable_loggers(enable_loggers)
    if disable_loggers is not None:
        _disable_loggers(disable_loggers)

    if not hasattr(LOGGER, 'is_complete'):
        setattr(LOGGER, 'is_complete', is_complete)
    if not hasattr(LOGGER, 'waitfor_complete'):
        setattr(LOGGER, 'waitfor_complete', waitfor_complete)

    return hndl

class _InterceptHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        # Get corresponding Loguru level
        level: str | int
        try:
            level = LOGGER.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller to get correct stack depth
        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1

        LOGGER.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage()
        )

        
# =============================================================================================
def _propogate_loggers(loggers: list[str]) -> bool:
    for lgr_name in [x for x in loggers if x.strip() != '']:
        logging_logger = logging.getLogger(lgr_name)
        logging_logger.handlers = []
        logging_logger.propagate = True
        LOGGER.trace(f'- propogate set for {lgr_name}')

    return True

def _enable_loggers(loggers: list[str]) -> bool:
    for lgr in [x for x in loggers if x.strip() != '']:
        LOGGER.enable(lgr)
        LOGGER.trace(f'- logging enabled for {lgr}')

    return True

def _disable_loggers(loggers: list[str]) -> bool:
    for lgr in [x for x in loggers if x.strip() != '']:
        LOGGER.disable(lgr)
        LOGGER.trace(f'- logging disabled for {lgr}')

    return True


# =============================================================================================
def _print_log_level_definitions():
    for lvl in ['TRACE','DEBUG','INFO','SUCCESS','WARNING','ERROR','CRITICAL']:
        LOGGER.log(lvl, LOGGER.level(lvl))
    while not LOGGER.complete():
        pass

def set_log_levels_brightness(on: bool = True):
    """
    Set brighness of console log messages

    Args:
        on (bool, optional): True messages are bold (bright), False messages are dimmer. Defaults to True.
    """
    for lvl in ['TRACE','DEBUG','INFO','SUCCESS','WARNING','ERROR','CRITICAL']:
        color = LOGGER.level(lvl).color
        if on and '<bold>' not in color:
            color = f'{color}<bold>'
        elif not on:
            color = color.replace('<bold>', '')
        LOGGER.level(lvl, color=color)

def is_complete() -> bool:
    """
    Check if any log output is pending.

    Returns:
        bool: True if enqueue buffer is empty, else False (i.e. pending log lines)
    """
    return LOGGER.complete()

def waitfor_complete():
    """
    If logger has enqueue set, wait for any pending log lines to 
    be written before continuing.
    """
    while not LOGGER.complete():
        time.sleep(.1)

# =============================================================================================
def disable_non_loguru_loggers():
    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    for lgr in loggers:
       logging.getLogger(lgr.name).disabled = True

def enable_non_loguru_loggers(logger_names: List[str]):
    for lgr in logger_names:
        logging.getLogger(lgr).disabled = False

def get_non_loguru_loggers() -> List[Dict[str, Dict]]:
    logger_list = []
    # loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]
    loggers = [logging.getLogger(name) for name in logging.RootLogger.manager.loggerDict]
    for lgr in loggers:
        lgr.level
        entry = {
            "name": lgr.name,
            "disabled": lgr.disabled, 
            "parent": lgr.parent.name,
            "filters": lgr.filters,
            "handlers": lgr.handlers,
            "level": lgr.level,
            "elevel": lgr.getEffectiveLevel(),
        }
        logger_list.append(entry)
    
    return logger_list

# =============================================================================================
# == Logging Decorators =======================================================================
# =============================================================================================
def logger_wraps(*, entry=True, exit=True, level="DEBUG"):
    """
    function decorator wrapper to log entry and exit

    When decorator enabled, messages will automatically be included in the the log:
    Example::    

        @logger_wraps()
        def foo(a, b, c):
            logger.info("Inside the function")
            return a * b * c 

    """
    def wrapper(func):
        name = func.__name__

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            logger_ = LOGGER.opt(depth=1)
            if entry:
                logger_.log(level, "Entering '{}' (args={}, kwargs={})", name, args, kwargs)
            result = func(*args, **kwargs)
            if exit:
                logger_.log(level, "Exiting '{}' (result={})", name, result)
            return result

        return wrapped

    return wrapper

def timer_wraps(*, level="DEBUG"):
    """
    function decorator wrapper to log function execution time

    Example::

        @timer_wraps(level='INFO')
        def foo(a,b,c):
          logger.info('inside the function')
          time.sleep(1)
          return a * b * c

    """
    def wrapper(func):
        name = func.__name__

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            """time_wrapper's doc string"""
            start = time.perf_counter()
            result = func(*args, **kwargs)
            time_elapsed = time.perf_counter() - start
            LOGGER.info(f"TIMER: Function: {name}, Time: {time_elapsed:.3f} seconds")
            return result
        return wrapped
    
    return wrapper


if __name__ == "__main__":
    import dt_tools.cli.demos.dt_foundation_demo as module
    module.demo()



"""
This module demonstrates the OSHelper features.

- OS Detection
- Elevated privilege detection and escalation

"""
from loguru import logger as LOGGER


import dt_tools.logger.logging_helper as lh
from dt_tools.os.os_helper import OSHelper
import sys

def demo():
    LOGGER.info('')
    LOGGER.info('-'*40)
    LOGGER.info('dt_misc_os_demo')
    LOGGER.info('-'*40)
    LOGGER.info('')
    LOGGER.info('Determine OS')
    LOGGER.info('------------')
    if OSHelper.is_raspberrypi():
        LOGGER.info('- Running on Raspberry Pi')
    if OSHelper.is_linux():
        LOGGER.success('- Running on Linux')
    if OSHelper.is_windows():
        LOGGER.success('- Running on Windows')

    LOGGER.info('')
    LOGGER.info('Check if Executable is available')
    LOGGER.info('--------------------------------')
    for pgm in ['gitx', 'git']:
        exe = OSHelper.is_executable_available(pgm)
        LOGGER.log("ERROR" if exe is None else "SUCCESS", f'  {pgm:8} : {exe}')

    if OSHelper.is_windows():
        LOGGER.info('')
        input('Press ENTER to Continue')
        LOGGER.info('')
        LOGGER.info('Administrator Check')
        LOGGER.info('-------------------')
        if OSHelper.is_windows_admin():
            LOGGER.info('')
            LOGGER.info('  ********************************************')
            LOGGER.info('  ** Windows admin privileges ARE in effect **')
            LOGGER.info('  ********************************************')
            LOGGER.info('')
        else:
            LOGGER.warning('  Not Admin, Elevate privileges')
            v_python = sys.executable
            args = [__file__]
            if OSHelper.elevate_to_admin(v_python, args):
                LOGGER.info('  - New prompt Shelled as admin')
            else:
                LOGGER.error('  - Unable to elevate. (User cancelled or error)')

        LOGGER.info('')
        LOGGER.info('Demo commplete.')            
        input('\nPress Enter to continue... ')

if __name__ == "__main__":
    lh.configure_logger(log_format=lh.DEFAULT_CONSOLE_LOGFMT, log_level="INFO")
    demo()
